# Copyright (c) 2025 Lakshya A Agrawal and the GEPA contributors
# https://github.com/gepa-ai/gepa

"""
GEPA Generic RAG Adapter Examples

This package contains examples demonstrating how to use GEPA's Generic RAG Adapter
with different vector stores and both local and cloud-based language models.
"""
